from typing import Final, Set

from moto.stepfunctions.parser.asl.component.component import Component


class Version(Component):
    _SUPPORTED_VERSIONS: Final[Set[str]] = {"1.0"}

    version: Final[str]

    def __init__(self, version: str):
        if version not in self._SUPPORTED_VERSIONS:
            raise ValueError(
                f"Version value '{version}' is not accepted. Supported Versions: {list(self._SUPPORTED_VERSIONS)}"
            )

        self.version = version



import abc
from typing import Final

from jsonpath_ng import parse

from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class MaxItemsDecl(EvalComponent, abc.ABC):
    @abc.abstractmethod
    def _get_value(self, env: Environment) -> int: ...

    def _eval_body(self, env: Environment) -> None:
        max_items: int = self._get_value(env=env)
        env.stack.append(max_items)


class MaxItems(MaxItemsDecl):
    """
    "MaxItems": Limits the number of data items passed to the Map state. For example, suppose that you provide a
    CSV file that contains 1000 rows and specify a limit of 100. Then, the interpreter passes only 100 rows to the
    Map state. The Map state processes items in sequential order, starting after the header row.
    Currently, you can specify a limit of up to 100,000,000
    """

    MAX_VALUE: Final[int] = 100_000_000

    max_items: Final[int]

    def __init__(self, max_items: int = MAX_VALUE):
        if max_items < 0 or max_items > MaxItems.MAX_VALUE:
            raise ValueError(
                f"MaxItems value MUST be a non-negative integer "
                f"non greater than '{MaxItems.MAX_VALUE}', got '{max_items}'."
            )
        self.max_items = max_items

    def _get_value(self, env: Environment) -> int:
        return self.max_items


class MaxItemsPath(MaxItemsDecl):
    """
    "MaxItemsPath": computes a MaxItems value equal to the reference path it points to.
    """

    def __init__(self, path: str):
        self.path: Final[str] = path

    def _get_value(self, env: Environment) -> int:
        input_expr = parse(self.path)
        max_items = input_expr.find(env.inp)
        return max_items



import json
from typing import Any
from urllib.parse import unquote

from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import BaseResponse

from .models import EKSBackend, eks_backends

DEFAULT_MAX_RESULTS = 100
DEFAULT_NEXT_TOKEN = ""


class EKSResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="eks")

    @property
    def eks_backend(self) -> EKSBackend:
        return eks_backends[self.current_account][self.region]

    def create_cluster(self) -> TYPE_RESPONSE:
        name = self._get_param("name")
        version = self._get_param("version")
        role_arn = self._get_param("roleArn")
        resources_vpc_config = self._get_param("resourcesVpcConfig")
        kubernetes_network_config = self._get_param("kubernetesNetworkConfig")
        logging = self._get_param("logging")
        client_request_token = self._get_param("clientRequestToken")
        tags = self._get_param("tags")
        encryption_config = self._get_param("encryptionConfig")

        cluster = self.eks_backend.create_cluster(
            name=name,
            version=version,
            role_arn=role_arn,
            resources_vpc_config=resources_vpc_config,
            kubernetes_network_config=kubernetes_network_config,
            logging=logging,
            client_request_token=client_request_token,
            tags=tags,
            encryption_config=encryption_config,
        )

        return 200, {}, json.dumps({"cluster": dict(cluster)})

    def create_fargate_profile(self) -> TYPE_RESPONSE:
        fargate_profile_name = self._get_param("fargateProfileName")
        cluster_name = self._get_param("name")
        pod_execution_role_arn = self._get_param("podExecutionRoleArn")
        subnets = self._get_param("subnets")
        selectors = self._get_param("selectors")
        client_request_token = self._get_param("clientRequestToken")
        tags = self._get_param("tags")

        fargate_profile = self.eks_backend.create_fargate_profile(
            fargate_profile_name=fargate_profile_name,
            cluster_name=cluster_name,
            pod_execution_role_arn=pod_execution_role_arn,
            subnets=subnets,
            selectors=selectors,
            client_request_token=client_request_token,
            tags=tags,
        )

        return 200, {}, json.dumps({"fargateProfile": dict(fargate_profile)})

    def create_nodegroup(self) -> TYPE_RESPONSE:
        cluster_name = self._get_param("name")
        nodegroup_name = self._get_param("nodegroupName")
        scaling_config = self._get_param("scalingConfig")
        disk_size = self._get_int_param("diskSize")
        subnets = self._get_param("subnets")
        instance_types = self._get_param("instanceTypes")
        ami_type = self._get_param("amiType")
        remote_access = self._get_param("remoteAccess")
        node_role = self._get_param("nodeRole")
        labels = self._get_param("labels")
        tags = self._get_param("tags")
        client_request_token = self._get_param("clientRequestToken")
        launch_template = self._get_param("launchTemplate")
        capacity_type = self._get_param("capacityType")
        version = self._get_param("version")
        release_version = self._get_param("releaseVersion")

        nodegroup = self.eks_backend.create_nodegroup(
            cluster_name=cluster_name,
            nodegroup_name=nodegroup_name,
            scaling_config=scaling_config,
            disk_size=disk_size,
            subnets=subnets,
            instance_types=instance_types,
            ami_type=ami_type,
            remote_access=remote_access,
            node_role=node_role,
            labels=labels,
            tags=tags,
            client_request_token=client_request_token,
            launch_template=launch_template,
            capacity_type=capacity_type,
            version=version,
            release_version=release_version,
        )

        return 200, {}, json.dumps({"nodegroup": dict(nodegroup)})

    def describe_cluster(self) -> TYPE_RESPONSE:
        name = self._get_param("name")

        cluster = self.eks_backend.describe_cluster(name=name)

        return 200, {}, json.dumps({"cluster": dict(cluster)})

    def describe_fargate_profile(self) -> TYPE_RESPONSE:
        cluster_name = self._get_param("name")
        fargate_profile_name = self._get_param("fargateProfileName")

        fargate_profile = self.eks_backend.describe_fargate_profile(
            cluster_name=cluster_name, fargate_profile_name=fargate_profile_name
        )
        return 200, {}, json.dumps({"fargateProfile": dict(fargate_profile)})

    def describe_nodegroup(self) -> TYPE_RESPONSE:
        cluster_name = self._get_param("name")
        nodegroup_name = self._get_param("nodegroupName")

        nodegroup = self.eks_backend.describe_nodegroup(
            cluster_name=cluster_name, nodegroup_name=nodegroup_name
        )

        return 200, {}, json.dumps({"nodegroup": dict(nodegroup)})

    def list_clusters(self) -> TYPE_RESPONSE:
        max_results = self._get_int_param("maxResults", DEFAULT_MAX_RESULTS)
        next_token = self._get_param("nextToken", DEFAULT_NEXT_TOKEN)

        clusters, next_token = self.eks_backend.list_clusters(
            max_results=max_results, next_token=next_token
        )

        return 200, {}, json.dumps(dict(clusters=clusters, nextToken=next_token))

    def list_fargate_profiles(self) -> TYPE_RESPONSE:
        cluster_name = self._get_param("name")
        max_results = self._get_int_param("maxResults", DEFAULT_MAX_RESULTS)
        next_token = self._get_param("nextToken", DEFAULT_NEXT_TOKEN)

        fargate_profile_names, next_token = self.eks_backend.list_fargate_profiles(
            cluster_name=cluster_name, max_results=max_results, next_token=next_token
        )

        return (
            200,
            {},
            json.dumps(
                dict(fargateProfileNames=fargate_profile_names, nextToken=next_token)
            ),
        )

    def list_nodegroups(self) -> TYPE_RESPONSE:
        cluster_name = self._get_param("name")
        max_results = self._get_int_param("maxResults", DEFAULT_MAX_RESULTS)
        next_token = self._get_param("nextToken", DEFAULT_NEXT_TOKEN)

        nodegroups, next_token = self.eks_backend.list_nodegroups(
            cluster_name=cluster_name, max_results=max_results, next_token=next_token
        )

        return 200, {}, json.dumps(dict(nodegroups=nodegroups, nextToken=next_token))

    def delete_cluster(self) -> TYPE_RESPONSE:
        name = self._get_param("name")

        cluster = self.eks_backend.delete_cluster(name=name)

        return 200, {}, json.dumps({"cluster": dict(cluster)})

    def delete_fargate_profile(self) -> TYPE_RESPONSE:
        cluster_name = self._get_param("name")
        fargate_profile_name = self._get_param("fargateProfileName")

        fargate_profile = self.eks_backend.delete_fargate_profile(
            cluster_name=cluster_name, fargate_profile_name=fargate_profile_name
        )

        return 200, {}, json.dumps({"fargateProfile": dict(fargate_profile)})

    def delete_nodegroup(self) -> TYPE_RESPONSE:
        cluster_name = self._get_param("name")
        nodegroup_name = self._get_param("nodegroupName")

        nodegroup = self.eks_backend.delete_nodegroup(
            cluster_name=cluster_name, nodegroup_name=nodegroup_name
        )

        return 200, {}, json.dumps({"nodegroup": dict(nodegroup)})

    def tags(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[return]
        self.setup_class(request, full_url, headers)
        if request.method == "GET":
            return self.list_tags_for_resource()
        if request.method == "POST":
            return self.tag_resource()
        if request.method == "DELETE":
            return self.untag_resource()

    def tag_resource(self) -> TYPE_RESPONSE:
        self.eks_backend.tag_resource(
            self._extract_arn_from_path(), self._get_param("tags")
        )

        return 200, {}, ""

    def untag_resource(self) -> TYPE_RESPONSE:
        self.eks_backend.untag_resource(
            self._extract_arn_from_path(), self._get_param("tagKeys")
        )

        return 200, {}, ""

    def list_tags_for_resource(self) -> TYPE_RESPONSE:
        tags = self.eks_backend.list_tags_for_resource(self._extract_arn_from_path())
        return 200, {}, json.dumps({"tags": tags})

    def _extract_arn_from_path(self) -> str:
        # /tags/arn_that_may_contain_a_slash
        path = unquote(self.path)
        return "/".join(path.split("/")[2:])



from typing import Final

from moto.stepfunctions.parser.asl.component.component import Component


class CauseDecl(Component):
    def __init__(self, cause: str):
        self.cause: Final[str] = cause



import inspect
from copy import deepcopy
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, TypeVar

from botocore.paginate import TokenDecoder, TokenEncoder

from moto.core.exceptions import InvalidToken

if TYPE_CHECKING:
    from typing_extensions import ParamSpec, Protocol

    P1 = ParamSpec("P1")
    P2 = ParamSpec("P2")
else:
    Protocol = object

T = TypeVar("T")


class GenericFunction(Protocol):
    def __call__(
        self, func: "Callable[P1, List[T]]"
    ) -> "Callable[P2, Tuple[List[T], Optional[str]]]": ...


def paginate(pagination_model: Dict[str, Any]) -> GenericFunction:
    def pagination_decorator(
        func: Callable[..., List[T]],
    ) -> Callable[..., Tuple[List[T], Optional[str]]]:
        @wraps(func)
        def pagination_wrapper(*args: Any, **kwargs: Any) -> Any:  # type: ignore
            method = func.__name__
            model = pagination_model
            pagination_config = model.get(method)
            if not pagination_config:
                raise ValueError(f"No pagination config for backend method: {method}")
            # Get the pagination arguments, to be used by the paginator
            next_token_name = pagination_config.get("input_token", "next_token")
            limit_name = pagination_config.get("limit_key")
            input_token = kwargs.get(next_token_name)
            limit = kwargs.get(limit_name, None)
            # Remove pagination arguments from our input kwargs
            # We need this to verify that our input kwargs are the same across invocations
            #   list_all(service="x")                   next_token = "a"
            #   list_all(service="x", next_token="a") ==> Works fine
            #   list_all(service="y", next_token="a") ==> Should throw an error, as the input_kwargs are different
            input_kwargs = deepcopy(kwargs)
            input_kwargs.pop(next_token_name, None)
            input_kwargs.pop(limit_name, None)
            fail_on_invalid_token = pagination_config.get("fail_on_invalid_token", True)
            paginator = Paginator(
                max_results=limit,
                max_results_default=pagination_config.get("limit_default"),
                starting_token=input_token,
                unique_attribute=pagination_config.get("unique_attribute"),
                param_values_to_check=input_kwargs,
                fail_on_invalid_token=fail_on_invalid_token,
            )

            # Determine which parameters to pass
            (arg_names, _, has_kwargs, _, _, _, _) = inspect.getfullargspec(func)
            # If the target-func expects `**kwargs`, we can pass everything
            if not has_kwargs:
                # If the target-function does not expect the next_token/limit, do not pass it
                if next_token_name not in arg_names:
                    kwargs.pop(next_token_name, None)
                if limit_name not in arg_names:
                    kwargs.pop(limit_name, None)

            results = func(*args, **kwargs)
            return paginator.paginate(results)

        return pagination_wrapper

    return pagination_decorator


class Paginator:
    def __init__(
        self,
        max_results: Any = None,
        max_results_default: Any = None,
        starting_token: Any = None,
        unique_attribute: Any = None,
        param_values_to_check: Any = None,
        fail_on_invalid_token: bool = True,
    ):
        self._max_results = max_results if max_results else max_results_default
        self._starting_token = starting_token
        self._unique_attributes = (
            unique_attribute
            if isinstance(unique_attribute, list)
            else [unique_attribute]
        )
        self._param_values_to_check = param_values_to_check
        self._fail_on_invalid_token = fail_on_invalid_token
        self._token_encoder = TokenEncoder()
        self._token_decoder = TokenDecoder()
        self._param_checksum = self._calculate_parameter_checksum()
        self._parsed_token = self._parse_starting_token()

    def _parse_starting_token(self) -> Optional[Dict[str, Any]]:
        if self._starting_token is None:
            return None
        # The starting token is a dict passed as a base64 encoded string.
        next_token = self._starting_token
        try:
            next_token = self._token_decoder.decode(next_token)
        except (ValueError, TypeError, UnicodeDecodeError):
            self._raise_exception_if_required(next_token)
            return None
        if next_token.get("parameterChecksum") != self._param_checksum:
            raise InvalidToken(f"Input inconsistent with page token: {str(next_token)}")
        return next_token

    def _raise_exception_if_required(self, token: Optional[str]) -> None:
        if self._fail_on_invalid_token:
            if isinstance(self._fail_on_invalid_token, type):
                # we need to raise a custom exception
                func_info = inspect.getfullargspec(self._fail_on_invalid_token)
                arg_names, _, _, _, _, _, _ = func_info
                # arg_names == [self] or [self, token_argument_that_can_have_any_name]
                requires_token_arg = len(arg_names) > 1
                if requires_token_arg:
                    raise self._fail_on_invalid_token(token)
                else:
                    raise self._fail_on_invalid_token()
            raise InvalidToken("Invalid token")

    def _calculate_parameter_checksum(self) -> int:
        def freeze(o: Any) -> Any:
            if not o:
                return None
            if isinstance(o, dict):
                return frozenset({k: freeze(v) for k, v in o.items()}.items())

            if isinstance(o, (list, tuple, set)):
                return tuple([freeze(v) for v in o])

            return o

        return hash(freeze(self._param_values_to_check))

    def _check_predicate(self, item: Any) -> bool:
        if self._parsed_token is None:
            return False
        unique_attributes = self._parsed_token["uniqueAttributes"]
        predicate_values = unique_attributes.split("|")
        for index, attr in enumerate(self._unique_attributes):
            curr_val = (
                item[attr] if isinstance(item, dict) else getattr(item, attr, None)
            )
            if not str(curr_val) == predicate_values[index]:
                return False
        return True

    def _build_next_token(self, next_item: Any) -> str:
        token_dict: Dict[str, Any] = {}
        if self._param_checksum:
            token_dict["parameterChecksum"] = self._param_checksum
        range_keys = []
        for attr in self._unique_attributes:
            if isinstance(next_item, dict):
                range_keys.append(str(next_item[attr]))
            else:
                range_keys.append(str(getattr(next_item, attr)))
        token_dict["uniqueAttributes"] = "|".join(range_keys)
        return self._token_encoder.encode(token_dict)

    def paginate(self, results: List[Any]) -> Tuple[List[Any], Optional[str]]:
        index_start = 0
        if self._starting_token:
            try:
                index_start = next(
                    index
                    for (index, result) in enumerate(results)
                    if self._check_predicate(result)
                )
            except StopIteration:
                if self._fail_on_invalid_token:
                    raise InvalidToken("Resource not found!")
                else:
                    return [], None

        index_end = index_start + self._max_results
        if index_end > len(results):
            index_end = len(results)

        results_page = results[index_start:index_end]

        next_token = None
        if results_page and index_end < len(results):
            last_resource_on_this_page = results[index_end]
            next_token = self._build_next_token(last_resource_on_this_page)
        return results_page, next_token



import json
import re
from typing import Any, Dict, Optional, Pattern, Tuple

from .exceptions import (
    DisabledCompatibilityVersioningException,
    GeneralGSRAlreadyExistsException,
    GeneralResourceNumberLimitExceededException,
    InvalidCompatibilityException,
    InvalidDataFormatException,
    InvalidNumberOfTagsException,
    InvalidRegistryIdBothParamsProvidedException,
    InvalidSchemaDefinitionException,
    InvalidSchemaIdBothParamsProvidedException,
    InvalidSchemaIdNotProvidedException,
    InvalidSchemaVersionIdProvidedWithOtherParamsException,
    InvalidSchemaVersionNumberBothParamsProvidedException,
    InvalidSchemaVersionNumberNotProvidedException,
    ParamValueContainsInvalidCharactersException,
    RegistryNotFoundException,
    ResourceNameTooLongException,
    SchemaNotFoundException,
    SchemaVersionMetadataLimitExceededException,
)
from .glue_schema_registry_constants import (
    ARN_PATTERN,
    DEFAULT_REGISTRY_NAME,
    DESCRIPTION_PATTERN,
    LATEST_VERSION,
    MAX_ARN_LENGTH,
    MAX_DESCRIPTION_LENGTH,
    MAX_REGISTRIES_ALLOWED,
    MAX_REGISTRY_NAME_LENGTH,
    MAX_SCHEMA_DEFINITION_LENGTH,
    MAX_SCHEMA_NAME_LENGTH,
    MAX_SCHEMA_VERSION_METADATA_ALLOWED,
    MAX_SCHEMA_VERSION_METADATA_LENGTH,
    MAX_SCHEMA_VERSIONS_ALLOWED,
    MAX_SCHEMAS_ALLOWED,
    MAX_TAGS_ALLOWED,
    METADATA_KEY,
    METADATA_VALUE,
    REGISTRY_ARN,
    REGISTRY_NAME,
    RESOURCE_NAME_PATTERN,
    SCHEMA_ARN,
    SCHEMA_DEFINITION,
    SCHEMA_NAME,
    SCHEMA_VERSION_ID,
    SCHEMA_VERSION_ID_PATTERN,
    SCHEMA_VERSION_METADATA_PATTERN,
    VERSION_NUMBER,
)


def validate_registry_name_pattern_and_length(param_value: str) -> None:
    validate_param_pattern_and_length(
        param_value,
        param_name="registryName",
        max_name_length=MAX_REGISTRY_NAME_LENGTH,
        pattern=RESOURCE_NAME_PATTERN,
    )


def validate_arn_pattern_and_length(param_value: str) -> None:
    validate_param_pattern_and_length(
        param_value,
        param_name="registryArn",
        max_name_length=MAX_ARN_LENGTH,
        pattern=ARN_PATTERN,
    )


def validate_description_pattern_and_length(param_value: str) -> None:
    validate_param_pattern_and_length(
        param_value,
        param_name="description",
        max_name_length=MAX_DESCRIPTION_LENGTH,
        pattern=DESCRIPTION_PATTERN,
    )


def validate_schema_name_pattern_and_length(param_value: str) -> None:
    validate_param_pattern_and_length(
        param_value,
        param_name="schemaName",
        max_name_length=MAX_SCHEMA_NAME_LENGTH,
        pattern=RESOURCE_NAME_PATTERN,
    )


def validate_schema_version_metadata_key_pattern_and_length(param_value: str) -> None:
    validate_param_pattern_and_length(
        param_value,
        param_name="key",
        max_name_length=MAX_SCHEMA_VERSION_METADATA_LENGTH,
        pattern=SCHEMA_VERSION_METADATA_PATTERN,
    )


def validate_schema_version_metadata_value_pattern_and_length(param_value: str) -> None:
    validate_param_pattern_and_length(
        param_value,
        param_name="value",
        max_name_length=MAX_SCHEMA_VERSION_METADATA_LENGTH,
        pattern=SCHEMA_VERSION_METADATA_PATTERN,
    )


def validate_param_pattern_and_length(
    param_value: str, param_name: str, max_name_length: int, pattern: Pattern[str]
) -> None:
    if len(param_value.encode("utf-8")) > max_name_length:
        raise ResourceNameTooLongException(param_name)

    if re.match(pattern, param_value) is None:
        raise ParamValueContainsInvalidCharactersException(param_name)


def validate_schema_definition(schema_definition: str, data_format: str) -> None:
    validate_schema_definition_length(schema_definition)
    if data_format in ["AVRO", "JSON"]:
        try:
            json.loads(schema_definition)
        except ValueError as err:
            raise InvalidSchemaDefinitionException(data_format, err)


def validate_schema_definition_length(schema_definition: str) -> None:
    if len(schema_definition) > MAX_SCHEMA_DEFINITION_LENGTH:
        param_name = SCHEMA_DEFINITION
        raise ResourceNameTooLongException(param_name)


def validate_schema_version_id_pattern(schema_version_id: str) -> None:
    if re.match(SCHEMA_VERSION_ID_PATTERN, schema_version_id) is None:
        raise ParamValueContainsInvalidCharactersException(SCHEMA_VERSION_ID)


def validate_number_of_tags(tags: Dict[str, str]) -> None:
    if len(tags) > MAX_TAGS_ALLOWED:
        raise InvalidNumberOfTagsException()


def validate_registry_id(
    registry_id: Dict[str, Any], registries: Dict[str, Any]
) -> str:
    if not registry_id:
        return DEFAULT_REGISTRY_NAME

    if registry_id.get(REGISTRY_NAME) and registry_id.get(REGISTRY_ARN):
        raise InvalidRegistryIdBothParamsProvidedException()

    if registry_id.get(REGISTRY_NAME):
        registry_name = registry_id.get(REGISTRY_NAME)
        validate_registry_name_pattern_and_length(registry_name)  # type: ignore

    elif registry_id.get(REGISTRY_ARN):
        registry_arn = registry_id.get(REGISTRY_ARN)
        validate_arn_pattern_and_length(registry_arn)  # type: ignore
        registry_name = registry_arn.split("/")[-1]  # type: ignore

    if registry_name != DEFAULT_REGISTRY_NAME and registry_name not in registries:
        if registry_id.get(REGISTRY_NAME):
            raise RegistryNotFoundException(
                resource="Registry",
                param_name=REGISTRY_NAME,
                param_value=registry_name,
            )
        if registry_id.get(REGISTRY_ARN):
            raise RegistryNotFoundException(
                resource="Registry",
                param_name=REGISTRY_ARN,
                param_value=registry_arn,
            )

    return registry_name  # type: ignore


def validate_registry_params(
    registries: Any,
    registry_name: str,
    description: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
) -> None:
    validate_registry_name_pattern_and_length(registry_name)

    if description:
        validate_description_pattern_and_length(description)

    if tags:
        validate_number_of_tags(tags)

    if len(registries) >= MAX_REGISTRIES_ALLOWED:
        raise GeneralResourceNumberLimitExceededException(resource="registries")

    if registry_name in registries:
        raise GeneralGSRAlreadyExistsException(
            resource="Registry",
            param_name=REGISTRY_NAME,
            param_value=registry_name,
        )


def validate_schema_id(
    schema_id: Dict[str, str], registries: Dict[str, Any]
) -> Tuple[str, str, Optional[str]]:
    schema_arn = schema_id.get(SCHEMA_ARN)
    registry_name = schema_id.get(REGISTRY_NAME)
    schema_name = schema_id.get(SCHEMA_NAME)
    if schema_arn:
        if registry_name or schema_name:
            raise InvalidSchemaIdBothParamsProvidedException()
        validate_arn_pattern_and_length(schema_arn)
        arn_components = schema_arn.split("/")
        schema_name = arn_components[-1]
        registry_name = arn_components[-2]

    else:
        if registry_name is None or schema_name is None:
            raise InvalidSchemaIdNotProvidedException()
        validate_registry_name_pattern_and_length(registry_name)
        validate_schema_name_pattern_and_length(schema_name)

    if (
        registry_name not in registries
        or schema_name not in registries[registry_name].schemas
    ):
        raise SchemaNotFoundException(schema_name, registry_name, schema_arn)

    return registry_name, schema_name, schema_arn


def validate_schema_params(
    registry: Any,
    schema_name: str,
    data_format: str,
    compatibility: str,
    schema_definition: str,
    num_schemas: int,
    description: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
) -> None:
    validate_schema_name_pattern_and_length(schema_name)

    if data_format not in ["AVRO", "JSON", "PROTOBUF"]:
        raise InvalidDataFormatException()

    if compatibility not in [
        "NONE",
        "DISABLED",
        "BACKWARD",
        "BACKWARD_ALL",
        "FORWARD",
        "FORWARD_ALL",
        "FULL",
        "FULL_ALL",
    ]:
        raise InvalidCompatibilityException()

    if description:
        validate_description_pattern_and_length(description)

    if tags:
        validate_number_of_tags(tags)

    validate_schema_definition(schema_definition, data_format)

    if num_schemas >= MAX_SCHEMAS_ALLOWED:
        raise GeneralResourceNumberLimitExceededException(resource="schemas")

    if schema_name in registry.schemas:
        raise GeneralGSRAlreadyExistsException(
            resource="Schema",
            param_name=SCHEMA_NAME,
            param_value=schema_name,
        )


def validate_register_schema_version_params(
    registry_name: str,
    schema_name: str,
    schema_arn: Optional[str],
    num_schema_versions: int,
    schema_definition: str,
    compatibility: str,
    data_format: str,
) -> None:
    if compatibility == "DISABLED":
        raise DisabledCompatibilityVersioningException(
            schema_name, registry_name, schema_arn
        )

    validate_schema_definition(schema_definition, data_format)

    if num_schema_versions >= MAX_SCHEMA_VERSIONS_ALLOWED:
        raise GeneralResourceNumberLimitExceededException(resource="schema versions")


def validate_schema_version_params(  # type: ignore[return]
    registries: Dict[str, Any],
    schema_id: Optional[Dict[str, Any]],
    schema_version_id: Optional[str],
    schema_version_number: Optional[Dict[str, Any]],
) -> Tuple[
    Optional[str],
    Optional[str],
    Optional[str],
    Optional[str],
    Optional[str],
    Optional[str],
]:
    if not schema_version_id and not schema_id and not schema_version_number:
        raise InvalidSchemaIdNotProvidedException()

    if schema_version_id and (schema_id or schema_version_number):
        raise InvalidSchemaVersionIdProvidedWithOtherParamsException()

    if schema_version_id:
        validate_schema_version_id_pattern(schema_version_id)

        # returns schema_version_id, registry_name, schema_name, schema_arn, version_number, latest_version
        return schema_version_id, None, None, None, None, None

    if schema_id and schema_version_number:
        registry_name, schema_name, schema_arn = validate_schema_id(
            schema_id, registries
        )
        version_number, latest_version = validate_schema_version_number(
            registries, registry_name, schema_name, schema_version_number
        )
        return (
            None,
            registry_name,
            schema_name,
            schema_arn,
            version_number,
            latest_version,
        )

    if not schema_id:
        raise InvalidSchemaIdNotProvidedException()

    if not schema_version_number:
        raise InvalidSchemaVersionNumberNotProvidedException()


def validate_schema_version_number(
    registries: Dict[str, Any],
    registry_name: str,
    schema_name: str,
    schema_version_number: Dict[str, str],
) -> Tuple[str, str]:
    latest_version = schema_version_number.get(LATEST_VERSION)
    version_number = schema_version_number.get(VERSION_NUMBER)
    schema = registries[registry_name].schemas[schema_name]
    if latest_version:
        if version_number:
            raise InvalidSchemaVersionNumberBothParamsProvidedException()
        return schema.latest_schema_version, latest_version

    return version_number, latest_version  # type: ignore


def validate_schema_version_metadata_pattern_and_length(
    metadata_key_value: Dict[str, str],
) -> Tuple[str, str]:
    metadata_key = metadata_key_value.get(METADATA_KEY)
    metadata_value = metadata_key_value.get(METADATA_VALUE)

    validate_schema_version_metadata_key_pattern_and_length(metadata_key)  # type: ignore
    validate_schema_version_metadata_value_pattern_and_length(metadata_value)  # type: ignore

    return metadata_key, metadata_value  # type: ignore[return-value]


def validate_number_of_schema_version_metadata_allowed(
    metadata: Dict[str, Any],
) -> None:
    num_metadata_key_value_pairs = 0
    for m in metadata.values():
        num_metadata_key_value_pairs += len(m)

    if num_metadata_key_value_pairs >= MAX_SCHEMA_VERSION_METADATA_ALLOWED:
        raise SchemaVersionMetadataLimitExceededException()


def get_schema_version_if_definition_exists(
    schema_versions: Any, data_format: str, schema_definition: str
) -> Optional[Dict[str, Any]]:
    if data_format in ["AVRO", "JSON"]:
        for schema_version in schema_versions:
            if json.loads(schema_definition) == json.loads(
                schema_version.schema_definition
            ):
                return schema_version.as_dict()
    else:
        for schema_version in schema_versions:
            if schema_definition == schema_version.schema_definition:
                return schema_version.as_dict()
    return None


def get_put_schema_version_metadata_response(
    schema_id: Dict[str, Any],
    schema_version_number: Optional[Dict[str, str]],
    schema_version_id: str,
    metadata_key_value: Dict[str, str],
) -> Dict[str, Any]:
    put_schema_version_metadata_response_dict: Dict[str, Any] = {}
    if schema_version_id:
        put_schema_version_metadata_response_dict[SCHEMA_VERSION_ID] = schema_version_id
    if schema_id:
        schema_arn = schema_id.get(SCHEMA_ARN)
        registry_name = schema_id.get(REGISTRY_NAME)
        schema_name = schema_id.get(SCHEMA_NAME)
        if schema_arn:
            put_schema_version_metadata_response_dict[SCHEMA_ARN] = schema_arn
        if registry_name:
            put_schema_version_metadata_response_dict[REGISTRY_NAME] = registry_name
        if schema_name:
            put_schema_version_metadata_response_dict[SCHEMA_NAME] = schema_name

    if schema_version_number:
        latest_version = schema_version_number.get(LATEST_VERSION)
        version_number = schema_version_number.get(VERSION_NUMBER)
        if latest_version:
            put_schema_version_metadata_response_dict[LATEST_VERSION] = latest_version
        else:
            put_schema_version_metadata_response_dict[LATEST_VERSION] = False

        if version_number:
            put_schema_version_metadata_response_dict[VERSION_NUMBER] = version_number
    else:
        put_schema_version_metadata_response_dict[LATEST_VERSION] = False
        put_schema_version_metadata_response_dict[VERSION_NUMBER] = 0

    metadata_key = metadata_key_value.get(METADATA_KEY)
    metadata_value = metadata_key_value.get(METADATA_VALUE)
    put_schema_version_metadata_response_dict[METADATA_KEY] = metadata_key
    put_schema_version_metadata_response_dict[METADATA_VALUE] = metadata_value

    return put_schema_version_metadata_response_dict


def delete_schema_response(
    schema_name: str, schema_arn: str, status: str
) -> Dict[str, Any]:
    return {
        "SchemaName": schema_name,
        "SchemaArn": schema_arn,
        "Status": status,
    }



from typing import Any, Final

from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payload_value import (
    PayloadValue,
)
from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payloadbinding.payload_binding import (
    PayloadBinding,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class PayloadBindingValue(PayloadBinding):
    def __init__(self, field: str, value: PayloadValue):
        super().__init__(field=field)
        self.value: Final[PayloadValue] = value

    def _eval_val(self, env: Environment) -> Any:
        self.value.eval(env)
        val: Any = env.stack.pop()
        return val



from typing import Any, Optional


def is_integer_between(
    x: int, mn: Optional[int] = None, mx: Optional[int] = None, optional: bool = False
) -> bool:
    if optional and x is None:
        return True
    try:
        if mn is not None and mx is not None:
            return int(x) >= mn and int(x) < mx
        elif mn is not None:
            return int(x) >= mn
        elif mx is not None:
            return int(x) < mx
        else:
            return True
    except ValueError:
        return False


def is_one_of(x: Any, choices: Any, optional: bool = False) -> bool:
    if optional and x is None:
        return True
    return x in choices



from moto.ec2.utils import add_tag_specification

from ._base_response import EC2BaseResponse


class CarrierGateway(EC2BaseResponse):
    def create_carrier_gateway(self) -> str:
        vpc_id = self._get_param("VpcId")
        tag_param = self._get_multi_param("TagSpecification")
        tags = add_tag_specification(tag_param)

        carrier_gateway = self.ec2_backend.create_carrier_gateway(
            vpc_id=vpc_id, tags=tags
        )
        template = self.response_template(CREATE_CARRIER_GATEWAY_RESPONSE)
        return template.render(carrier_gateway=carrier_gateway)

    def delete_carrier_gateway(self) -> str:
        carrier_gateway_id = self._get_param("CarrierGatewayId")

        carrier_gateway = self.ec2_backend.delete_carrier_gateway(carrier_gateway_id)
        template = self.response_template(DELETE_CARRIER_GATEWAY_RESPONSE)
        return template.render(carrier_gateway=carrier_gateway)

    def describe_carrier_gateways(self) -> str:
        carrier_gateway_ids = self._get_multi_param("CarrierGatewayId")
        filters = self._filters_from_querystring()

        carrier_gateways = self.ec2_backend.describe_carrier_gateways(
            carrier_gateway_ids, filters
        )
        template = self.response_template(DESCRIBE_CARRIER_GATEWAYS_RESPONSE)
        return template.render(carrier_gateways=carrier_gateways)


CREATE_CARRIER_GATEWAY_RESPONSE = """<CreateCarrierGatewayResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
    <requestId>c617595f-6c29-4a00-a941-example</requestId>
    <carrierGateway>
        <state>{{ carrier_gateway.state }}</state>
        <vpcId>{{ carrier_gateway.vpc_id }}</vpcId>
        <carrierGatewayId>{{ carrier_gateway.id }}</carrierGatewayId>
        <ownerId>{{ carrier_gateway.owner_id }}</ownerId>
        <tagSet>
        {% for tag in carrier_gateway.get_tags() %}
            <item>
                <key>{{ tag.key }}</key>
                <value>{{ tag.value }}</value>
            </item>
        {% endfor %}
        </tagSet>
    </carrierGateway>
</CreateCarrierGatewayResponse>
"""

DELETE_CARRIER_GATEWAY_RESPONSE = """<DeleteCarrierGatewayResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
    <requestId>c617595f-6c29-4a00-a941-example</requestId>
    <carrierGateway>
        <state>{{ carrier_gateway.state }}</state>
        <vpcId>{{ carrier_gateway.vpc_id }}</vpcId>
        <carrierGatewayId>{{ carrier_gateway.id }}</carrierGatewayId>
        <ownerId>{{ carrier_gateway.owner_id }}</ownerId>
        <tagSet>
        {% for tag in carrier_gateway.get_tags() %}
            <item>
                <key>{{ tag.key }}</key>
                <value>{{ tag.value }}</value>
            </item>
        {% endfor %}
        </tagSet>
    </carrierGateway>
</DeleteCarrierGatewayResponse>
"""

DESCRIBE_CARRIER_GATEWAYS_RESPONSE = """<DescribeCarrierGatewaysResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
    <requestId>151283df-f7dc-4317-89b4-01c9888b1d45</requestId>
    <carrierGatewaySet>
    {% for carrier_gateway in carrier_gateways %}
        <item>
            <state>{{ carrier_gateway.state }}</state>
            <vpcId>{{ carrier_gateway.vpc_id }}</vpcId>
            <carrierGatewayId>{{ carrier_gateway.id }}</carrierGatewayId>
            <ownerId>{{ carrier_gateway.owner_id }}</ownerId>
            <tagSet>
            {% for tag in carrier_gateway.get_tags() %}
                <item>
                    <key>{{ tag.key }}</key>
                    <value>{{ tag.value }}</value>
                </item>
            {% endfor %}
            </tagSet>
        </item>
    {% endfor %}
    </carrierGatewaySet>
</DescribeCarrierGatewaysResponse>
"""



from __future__ import annotations

import abc
import copy
from typing import Any, Dict, Final, List, Optional, Tuple

from botocore.model import StructureShape

from moto.stepfunctions.parser.api import (
    HistoryEventExecutionDataDetails,
    HistoryEventType,
    TaskScheduledEventDetails,
    TaskStartedEventDetails,
    TaskSucceededEventDetails,
    TaskTimedOutEventDetails,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ResourceRuntimePart,
    ServiceResource,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.state_task import (
    StateTask,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str
from moto.stepfunctions.parser.utils import camel_to_snake_case, snake_to_camel_case


class StateTaskService(StateTask, abc.ABC):
    resource: ServiceResource

    _SERVICE_NAME_SFN_TO_BOTO_OVERRIDES: Final[Dict[str, str]] = {
        "sfn": "stepfunctions",
        "states": "stepfunctions",
    }

    def _get_sfn_resource(self) -> str:
        return self.resource.api_action

    def _get_sfn_resource_type(self) -> str:
        return self.resource.service_name

    def _get_timed_out_failure_event(self) -> FailureEvent:
        return FailureEvent(
            error_name=StatesErrorName(typ=StatesErrorNameType.StatesTimeout),
            event_type=HistoryEventType.TaskTimedOut,
            event_details=EventDetails(
                taskTimedOutEventDetails=TaskTimedOutEventDetails(
                    resourceType=self._get_sfn_resource_type(),
                    resource=self._get_sfn_resource(),
                    error=StatesErrorNameType.StatesTimeout.to_name(),
                )
            ),
        )

    def _to_boto_args(self, parameters: dict, structure_shape: StructureShape) -> None:
        shape_members = structure_shape.members
        norm_member_binds: Dict[str, Tuple[str, Optional[StructureShape]]] = {
            camel_to_snake_case(member_key): (
                member_key,
                member_value if isinstance(member_value, StructureShape) else None,
            )
            for member_key, member_value in shape_members.items()
        }
        parameters_bind_keys: List[str] = list(parameters.keys())
        for parameter_key in parameters_bind_keys:
            norm_parameter_key = camel_to_snake_case(parameter_key)
            norm_member_bind: Optional[Tuple[str, Optional[StructureShape]]] = (
                norm_member_binds.get(norm_parameter_key)
            )
            if norm_member_bind is not None:
                norm_member_bind_key, norm_member_bind_shape = norm_member_bind
                parameter_value = parameters.pop(parameter_key)
                if norm_member_bind_shape is not None:
                    self._to_boto_args(parameter_value, norm_member_bind_shape)
                parameters[norm_member_bind_key] = parameter_value

    @staticmethod
    def _to_sfn_cased(member_key: str) -> str:
        # Normalise the string to snake case, e.g. "HelloWorld_hello__world" -> "hello_world_hello_world"
        norm_member_key = camel_to_snake_case(member_key)
        # Normalise the snake case to camel case, e.g. "hello_world_hello_world" -> "HelloWorldHelloWorld"
        norm_member_key = snake_to_camel_case(norm_member_key)
        return norm_member_key

    def _from_boto_response(
        self, response: Any, structure_shape: StructureShape
    ) -> None:
        if not isinstance(response, dict):
            return

        shape_members = structure_shape.members
        response_bind_keys: List[str] = list(response.keys())
        for response_key in response_bind_keys:
            norm_response_key = self._to_sfn_cased(response_key)
            if response_key in shape_members:
                shape_member = shape_members[response_key]

                response_value = response.pop(response_key)
                if isinstance(shape_member, StructureShape):
                    self._from_boto_response(response_value, shape_member)
                response[norm_response_key] = response_value

    def _get_boto_service_name(self, boto_service_name: Optional[str] = None) -> str:
        api_name = boto_service_name or self.resource.api_name
        return self._SERVICE_NAME_SFN_TO_BOTO_OVERRIDES.get(api_name, api_name)

    def _get_boto_service_action(
        self, service_action_name: Optional[str] = None
    ) -> str:
        api_action = service_action_name or self.resource.api_action
        return camel_to_snake_case(api_action)

    def _normalise_parameters(
        self,
        parameters: dict,
        boto_service_name: Optional[str] = None,
        service_action_name: Optional[str] = None,
    ) -> None:
        pass

    def _normalise_response(
        self,
        response: Any,
        boto_service_name: Optional[str] = None,
        service_action_name: Optional[str] = None,
    ) -> None:
        pass

    @abc.abstractmethod
    def _eval_service_task(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
    ): ...

    def _before_eval_execution(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        raw_parameters: dict,
    ) -> None:
        parameters_str = to_json_str(raw_parameters)

        scheduled_event_details = TaskScheduledEventDetails(
            resource=self._get_sfn_resource(),
            resourceType=self._get_sfn_resource_type(),
            region=resource_runtime_part.region,
            parameters=parameters_str,
        )
        if not self.timeout.is_default_value():
            self.timeout.eval(env=env)
            timeout_seconds = env.stack.pop()
            scheduled_event_details["timeoutInSeconds"] = timeout_seconds
        if self.heartbeat is not None:
            self.heartbeat.eval(env=env)
            heartbeat_seconds = env.stack.pop()
            scheduled_event_details["heartbeatInSeconds"] = heartbeat_seconds
        env.event_history.add_event(
            context=env.event_history_context,
            hist_type_event=HistoryEventType.TaskScheduled,
            event_detail=EventDetails(
                taskScheduledEventDetails=scheduled_event_details
            ),
        )

        env.event_history.add_event(
            context=env.event_history_context,
            hist_type_event=HistoryEventType.TaskStarted,
            event_detail=EventDetails(
                taskStartedEventDetails=TaskStartedEventDetails(
                    resource=self._get_sfn_resource(),
                    resourceType=self._get_sfn_resource_type(),
                )
            ),
        )

    def _after_eval_execution(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
    ) -> None:
        output = env.stack[-1]
        env.event_history.add_event(
            context=env.event_history_context,
            hist_type_event=HistoryEventType.TaskSucceeded,
            event_detail=EventDetails(
                taskSucceededEventDetails=TaskSucceededEventDetails(
                    resource=self._get_sfn_resource(),
                    resourceType=self._get_sfn_resource_type(),
                    output=to_json_str(output),
                    outputDetails=HistoryEventExecutionDataDetails(truncated=False),
                )
            ),
        )

    def _eval_execution(self, env: Environment) -> None:
        self.resource.eval(env=env)
        resource_runtime_part: ResourceRuntimePart = env.stack.pop()

        raw_parameters = self._eval_parameters(env=env)

        self._before_eval_execution(
            env=env,
            resource_runtime_part=resource_runtime_part,
            raw_parameters=raw_parameters,
        )

        normalised_parameters = copy.deepcopy(raw_parameters)
        self._normalise_parameters(normalised_parameters)

        self._eval_service_task(
            env=env,
            resource_runtime_part=resource_runtime_part,
            normalised_parameters=normalised_parameters,
        )

        output_value = env.stack[-1]
        self._normalise_response(output_value)

        self._after_eval_execution(
            env=env,
            resource_runtime_part=resource_runtime_part,
            normalised_parameters=normalised_parameters,
        )



import json
from typing import Dict, Final, Optional, Set

from botocore.exceptions import ClientError

from moto.stepfunctions.parser.api import (
    DescribeExecutionOutput,
    ExecutionStatus,
    HistoryEventType,
    TaskFailedEventDetails,
)
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ResourceRuntimePart,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_callback import (
    StateTaskServiceCallback,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.boto_client import boto_client_for
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str


class StateTaskServiceSfn(StateTaskServiceCallback):
    _SUPPORTED_API_PARAM_BINDINGS: Final[Dict[str, Set[str]]] = {
        "startexecution": {"Input", "Name", "StateMachineArn"}
    }

    def _get_supported_parameters(self) -> Optional[Set[str]]:
        return self._SUPPORTED_API_PARAM_BINDINGS.get(self.resource.api_action.lower())

    def _from_error(self, env: Environment, ex: Exception) -> FailureEvent:
        if isinstance(ex, ClientError):
            error_code = ex.response["Error"]["Code"]
            error_name: str = f"StepFunctions.{error_code}Exception"
            error_cause_details = [
                "Service: AWSStepFunctions",
                f"Status Code: {ex.response['ResponseMetadata']['HTTPStatusCode']}",
                f"Error Code: {error_code}",
                f"Request ID: {ex.response['ResponseMetadata']['RequestId']}",
                "Proxy: null",  # TODO: investigate this proxy value.
            ]
            if "HostId" in ex.response["ResponseMetadata"]:
                error_cause_details.append(
                    f'Extended Request ID: {ex.response["ResponseMetadata"]["HostId"]}'
                )
            error_cause: str = (
                f"{ex.response['Error']['Message']} ({'; '.join(error_cause_details)})"
            )
            return FailureEvent(
                error_name=CustomErrorName(error_name),
                event_type=HistoryEventType.TaskFailed,
                event_details=EventDetails(
                    taskFailedEventDetails=TaskFailedEventDetails(
                        error=error_name,
                        cause=error_cause,
                        resource=self._get_sfn_resource(),
                        resourceType=self._get_sfn_resource_type(),
                    )
                ),
            )
        return super()._from_error(env=env, ex=ex)

    def _normalise_parameters(
        self,
        parameters: dict,
        boto_service_name: Optional[str] = None,
        service_action_name: Optional[str] = None,
    ) -> None:
        if service_action_name is None:
            if self._get_boto_service_action() == "start_execution":
                optional_input = parameters.get("Input")
                if not isinstance(optional_input, str):
                    # AWS Sfn's documentation states:
                    # If you don't include any JSON input data, you still must include the two braces.
                    if optional_input is None:
                        optional_input = {}
                    parameters["Input"] = to_json_str(
                        optional_input, separators=(",", ":")
                    )
        super()._normalise_parameters(
            parameters=parameters,
            boto_service_name=boto_service_name,
            service_action_name=service_action_name,
        )

    @staticmethod
    def _sync2_api_output_of(typ: type, value: json) -> None:
        def _replace_with_json_if_str(key: str) -> None:
            inner_value = value.get(key)
            if isinstance(inner_value, str):
                value[key] = json.loads(inner_value)

        if typ == DescribeExecutionOutput:
            _replace_with_json_if_str("input")
            _replace_with_json_if_str("output")

    def _eval_service_task(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
    ):
        service_name = self._get_boto_service_name()
        api_action = self._get_boto_service_action()
        sfn_client = boto_client_for(
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            service=service_name,
        )
        response = getattr(sfn_client, api_action)(**normalised_parameters)
        response.pop("ResponseMetadata", None)
        env.stack.append(response)

    def _sync_to_start_machine(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        sync2_response: bool,
    ) -> None:
        sfn_client = boto_client_for(
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            service="stepfunctions",
        )
        submission_output: dict = env.stack.pop()
        execution_arn: str = submission_output["ExecutionArn"]

        def _has_terminated() -> Optional[dict]:
            describe_execution_output = sfn_client.describe_execution(
                executionArn=execution_arn
            )
            execution_status: ExecutionStatus = describe_execution_output["status"]

            if execution_status != ExecutionStatus.RUNNING:
                if sync2_response:
                    self._sync2_api_output_of(
                        typ=DescribeExecutionOutput, value=describe_execution_output
                    )
                self._normalise_response(
                    response=describe_execution_output,
                    service_action_name="describe_execution",
                )
                if execution_status == ExecutionStatus.SUCCEEDED:
                    return describe_execution_output
                else:
                    raise FailureEventException(
                        FailureEvent(
                            error_name=StatesErrorName(
                                typ=StatesErrorNameType.StatesTaskFailed
                            ),
                            event_type=HistoryEventType.TaskFailed,
                            event_details=EventDetails(
                                taskFailedEventDetails=TaskFailedEventDetails(
                                    resource=self._get_sfn_resource(),
                                    resourceType=self._get_sfn_resource_type(),
                                    error=StatesErrorNameType.StatesTaskFailed.to_name(),
                                    cause=to_json_str(describe_execution_output),
                                )
                            ),
                        )
                    )
            return None

        termination_output: Optional[dict] = None
        while env.is_running() and not termination_output:
            termination_output: Optional[dict] = _has_terminated()

        env.stack.append(termination_output)

    def _sync(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
    ) -> None:
        if self._get_boto_service_action() == "start_execution":
            return self._sync_to_start_machine(
                env=env,
                resource_runtime_part=resource_runtime_part,
                sync2_response=False,
            )
        else:
            super()._sync(
                env=env,
                resource_runtime_part=resource_runtime_part,
                normalised_parameters=normalised_parameters,
            )

    def _sync2(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
    ) -> None:
        if self._get_boto_service_action() == "start_execution":
            return self._sync_to_start_machine(
                env=env,
                resource_runtime_part=resource_runtime_part,
                sync2_response=True,
            )
        else:
            super()._sync2(
                env=env,
                resource_runtime_part=resource_runtime_part,
                normalised_parameters=normalised_parameters,
            )



from typing import Any, Final, Optional, TypedDict

from moto.moto_api._internal import mock_random


class Execution(TypedDict):
    Id: str
    Input: Optional[dict]
    Name: str
    RoleArn: str
    StartTime: str  # Format: ISO 8601.


class State(TypedDict):
    EnteredTime: str  # Format: ISO 8601.
    Name: str
    RetryCount: int


class StateMachine(TypedDict):
    Id: str
    Name: str


class Task(TypedDict):
    Token: str


class Item(TypedDict):
    # Contains the index number for the array item that is being currently processed.
    Index: int
    # Contains the array item being processed.
    Value: Optional[Any]


class Map(TypedDict):
    Item: Item


class ContextObject(TypedDict):
    Execution: Execution
    State: Optional[State]
    StateMachine: StateMachine
    Task: Optional[Task]  # Null if the Parameters field is outside a task state.
    Map: Optional[Map]  # Only available when processing a Map state.


class ContextObjectManager:
    context_object: Final[ContextObject]

    def __init__(self, context_object: ContextObject):
        self.context_object = context_object

    def update_task_token(self) -> str:
        new_token = str(mock_random.uuid4())
        self.context_object["Task"] = Task(Token=new_token)
        return new_token


class ContextObjectInitData(TypedDict):
    Execution: Execution
    StateMachine: StateMachine



from moto.core.exceptions import JsonRESTError


class ResourceNotFoundError(JsonRESTError):
    def __init__(self, message: str):
        super().__init__(error_type="ResourceNotFoundException", message=message)


class InvalidNameException(JsonRESTError):
    message = "1 validation error detected: Value '{}' at 'identityPoolName' failed to satisfy constraint: Member must satisfy regular expression pattern: [\\w\\s+=,.@-]+"

    def __init__(self, name: str):
        msg = InvalidNameException.message.format(name)
        super().__init__(error_type="ValidationException", message=msg)



from moto.core.responses import BaseResponse


class IPAddresses(BaseResponse):
    def assign_private_ip_addresses(self) -> str:
        self.error_on_dryrun()

        raise NotImplementedError(
            "IPAddresses.assign_private_ip_addresses is not yet implemented"
        )

    def unassign_private_ip_addresses(self) -> str:
        self.error_on_dryrun()

        raise NotImplementedError(
            "IPAddresses.unassign_private_ip_addresses is not yet implemented"
        )



from __future__ import annotations

from typing import Final

from moto.stepfunctions.parser.asl.component.common.error_name.error_name import (
    ErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)


class StatesErrorName(ErrorName):
    def __init__(self, typ: StatesErrorNameType):
        super().__init__(error_name=typ.to_name())
        self.typ: Final[StatesErrorNameType] = typ

    @classmethod
    def from_name(cls, error_name: str) -> StatesErrorName:
        error_name_type: StatesErrorNameType = StatesErrorNameType.from_name(error_name)
        return cls(typ=error_name_type)



"""workspaces base URL and path."""

from .responses import WorkSpacesResponse

url_bases = [
    r"https?://workspaces\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/$": WorkSpacesResponse.dispatch,
}



from .responses import DataBrewResponse

url_bases = [r"https?://databrew\.(.+)\.amazonaws.com"]

url_paths = {
    "{0}/recipeVersions$": DataBrewResponse.dispatch,
    "{0}/recipes$": DataBrewResponse.dispatch,
    "{0}/recipes/(?P<recipe_name>[^/]+)$": DataBrewResponse.dispatch,
    "{0}/recipes/(?P<recipe_name>[^/]+)/recipeVersion/(?P<recipe_version>[^/]+)": DataBrewResponse.dispatch,
    "{0}/recipes/(?P<recipe_name>[^/]+)/publishRecipe$": DataBrewResponse.dispatch,
    "{0}/rulesets$": DataBrewResponse.dispatch,
    "{0}/rulesets/(?P<ruleset_name>[^/]+)$": DataBrewResponse.dispatch,
    "{0}/datasets$": DataBrewResponse.dispatch,
    "{0}/datasets/(?P<dataset_name>[^/]+)$": DataBrewResponse.dispatch,
    "{0}/jobs$": DataBrewResponse.dispatch,
    "{0}/jobs/(?P<job_name>[^/]+)$": DataBrewResponse.dispatch,
    "{0}/profileJobs$": DataBrewResponse.dispatch,
    "{0}/recipeJobs$": DataBrewResponse.dispatch,
    "{0}/profileJobs/(?P<job_name>[^/]+)$": DataBrewResponse.dispatch,
    "{0}/recipeJobs/(?P<job_name>[^/]+)$": DataBrewResponse.dispatch,
}



import base64
from datetime import datetime
from typing import Any, List, Optional

from .exceptions import InvalidArgumentError

encode_method = base64.encodebytes
decode_method = base64.decodebytes


PAGINATION_MODEL = {
    "list_shards": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 10000,
        "unique_attribute": "ShardId",
        "fail_on_invalid_token": False,
    },
}


PAGINATION_MODEL = {
    "list_shards": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 10000,
        "unique_attribute": "ShardId",
        "fail_on_invalid_token": False,
    },
}


def compose_new_shard_iterator(
    stream_name: Optional[str],
    shard: Any,
    shard_iterator_type: str,
    starting_sequence_number: int,
    at_timestamp: datetime,
) -> str:
    if shard_iterator_type == "AT_SEQUENCE_NUMBER":
        last_sequence_id = int(starting_sequence_number) - 1
    elif shard_iterator_type == "AFTER_SEQUENCE_NUMBER":
        last_sequence_id = int(starting_sequence_number)
    elif shard_iterator_type == "TRIM_HORIZON":
        last_sequence_id = 0
    elif shard_iterator_type == "LATEST":
        last_sequence_id = shard.get_max_sequence_number()
    elif shard_iterator_type == "AT_TIMESTAMP":
        last_sequence_id = shard.get_sequence_number_at(at_timestamp)
    else:
        raise InvalidArgumentError(f"Invalid ShardIteratorType: {shard_iterator_type}")
    return compose_shard_iterator(stream_name, shard, last_sequence_id)


def compose_shard_iterator(
    stream_name: Optional[str], shard: Any, last_sequence_id: int
) -> str:
    return encode_method(
        f"{stream_name}:{shard.shard_id}:{last_sequence_id}".encode("utf-8")
    ).decode("utf-8")


def decompose_shard_iterator(shard_iterator: str) -> List[str]:
    return decode_method(shard_iterator.encode("utf-8")).decode("utf-8").split(":")



from __future__ import annotations

import abc
import datetime
import json
from collections import OrderedDict
from typing import Dict, Final, Optional

from moto.moto_api._internal import mock_random
from moto.stepfunctions.parser.api import (
    Definition,
    DescribeStateMachineOutput,
    LoggingConfiguration,
    Name,
    RevisionId,
    StateMachineListItem,
    StateMachineStatus,
    StateMachineType,
    StateMachineVersionListItem,
    Tag,
    TagKeyList,
    TagList,
    TracingConfiguration,
    ValidationException,
)


class StateMachineInstance:
    name: Name
    arn: str
    revision_id: Optional[RevisionId]
    definition: Definition
    role_arn: str
    create_date: datetime.datetime
    sm_type: StateMachineType
    logging_config: Optional[LoggingConfiguration]
    tags: Optional[TagList]
    tracing_config: Optional[TracingConfiguration]

    def __init__(
        self,
        name: Name,
        arn: str,
        definition: Definition,
        role_arn: str,
        create_date: Optional[datetime.datetime] = None,
        sm_type: Optional[StateMachineType] = None,
        logging_config: Optional[LoggingConfiguration] = None,
        tags: Optional[TagList] = None,
        tracing_config: Optional[TracingConfiguration] = None,
    ):
        self.name = name
        self.arn = arn
        self.revision_id = None
        self.definition = definition
        self.role_arn = role_arn
        self.create_date = create_date or datetime.datetime.now(
            tz=datetime.timezone.utc
        )
        self.sm_type = sm_type or StateMachineType.STANDARD
        self.logging_config = logging_config
        self.tags = tags
        self.tracing_config = tracing_config

    def describe(self) -> DescribeStateMachineOutput:
        describe_output = DescribeStateMachineOutput(
            stateMachineArn=self.arn,
            name=self.name,
            status=StateMachineStatus.ACTIVE,
            definition=self.definition,
            roleArn=self.role_arn,
            type=self.sm_type,
            creationDate=self.create_date,
            loggingConfiguration=self.logging_config,
        )
        if self.revision_id:
            describe_output["revisionId"] = self.revision_id
        return describe_output

    @abc.abstractmethod
    def itemise(self): ...


class TagManager:
    _tags: Final[Dict[str, Optional[str]]]

    def __init__(self):
        self._tags = OrderedDict()

    @staticmethod
    def _validate_key_value(key: str) -> None:
        if not key:
            raise ValidationException()

    @staticmethod
    def _validate_tag_value(value: str) -> None:
        if value is None:
            raise ValidationException()

    def add_all(self, tags: TagList) -> None:
        for tag in tags:
            tag_key = tag["key"]
            tag_value = tag["value"]
            self._validate_key_value(key=tag_key)
            self._validate_tag_value(value=tag_value)
            self._tags[tag_key] = tag_value

    def remove_all(self, keys: TagKeyList):
        for key in keys:
            self._validate_key_value(key=key)
            self._tags.pop(key, None)

    def to_tag_list(self) -> TagList:
        tag_list = list()
        for key, value in self._tags.items():
            tag_list.append(Tag(key=key, value=value))
        return tag_list


class StateMachineRevision(StateMachineInstance):
    _next_version_number: int
    versions: Final[Dict[RevisionId, str]]
    tag_manager: Final[TagManager]

    def __init__(
        self,
        name: Name,
        arn: str,
        definition: Definition,
        role_arn: str,
        create_date: Optional[datetime.datetime] = None,
        sm_type: Optional[StateMachineType] = None,
        logging_config: Optional[LoggingConfiguration] = None,
        tags: Optional[TagList] = None,
        tracing_config: Optional[TracingConfiguration] = None,
    ):
        super().__init__(
            name,
            arn,
            definition,
            role_arn,
            create_date,
            sm_type,
            logging_config,
            tags,
            tracing_config,
        )
        self.versions = dict()
        self._version_number = 0
        self.tag_manager = TagManager()

    def create_revision(
        self, definition: Optional[str], role_arn: Optional[str]
    ) -> Optional[RevisionId]:
        update_definition = definition and json.loads(definition) != json.loads(
            self.definition
        )
        if update_definition:
            self.definition = definition

        update_role_arn = role_arn and role_arn != self.role_arn
        if update_role_arn:
            self.role_arn = role_arn

        if any([update_definition, update_role_arn]):
            self.revision_id = str(mock_random.uuid4())

        return self.revision_id

    def create_version(
        self, description: Optional[str]
    ) -> Optional[StateMachineVersion]:
        if self.revision_id not in self.versions:
            self._version_number += 1
            version = StateMachineVersion(
                self, version=self._version_number, description=description
            )
            self.versions[self.revision_id] = version.arn

            return version
        return None

    def delete_version(self, state_machine_version_arn: str) -> None:
        source_revision_id = None
        for revision_id, version_arn in self.versions.items():
            if version_arn == state_machine_version_arn:
                source_revision_id = revision_id
                break
        self.versions.pop(source_revision_id, None)

    def itemise(self) -> StateMachineListItem:
        return StateMachineListItem(
            stateMachineArn=self.arn,
            name=self.name,
            type=self.sm_type,
            creationDate=self.create_date,
        )


class StateMachineVersion(StateMachineInstance):
    source_arn: str
    version: int
    description: Optional[str]

    def __init__(
        self,
        state_machine_revision: StateMachineRevision,
        version: int,
        description: Optional[str],
    ):
        version_arn = f"{state_machine_revision.arn}:{version}"
        super().__init__(
            name=state_machine_revision.name,
            arn=version_arn,
            definition=state_machine_revision.definition,
            role_arn=state_machine_revision.role_arn,
            create_date=datetime.datetime.now(tz=datetime.timezone.utc),
            sm_type=state_machine_revision.sm_type,
            logging_config=state_machine_revision.logging_config,
            tags=state_machine_revision.tags,
            tracing_config=state_machine_revision.tracing_config,
        )
        self.source_arn = state_machine_revision.arn
        self.revision_id = state_machine_revision.revision_id
        self.version = version
        self.description = description

    def describe(self) -> DescribeStateMachineOutput:
        describe_output: DescribeStateMachineOutput = super().describe()
        if self.description:
            describe_output["description"] = self.description
        return describe_output

    def itemise(self) -> StateMachineVersionListItem:
        return StateMachineVersionListItem(
            stateMachineVersionArn=self.arn, creationDate=self.create_date
        )



import logging
import sys

log_format = "%(levelname)s %(asctime)s - %(message)s"
logging.basicConfig(stream=sys.stdout, format=log_format)
logger = logging.getLogger("MOTO_PROXY")
logger.setLevel(logging.INFO)


def with_color(color: int, text: object) -> str:
    return f"\x1b[{color}m{text}\x1b[0m"


def info(msg: object) -> None:
    logger.info(msg)


def debug(msg: object) -> None:
    logger.debug(msg)


def error(msg: object) -> None:
    logger.error(msg)



from typing import TYPE_CHECKING, Any, Dict, Optional

from moto.core.common_models import BaseModel
from moto.core.utils import unix_time, utcnow
from moto.moto_api._internal import mock_random

from ..exceptions import SWFWorkflowExecutionClosedError
from .timeout import Timeout

if TYPE_CHECKING:
    from .activity_type import ActivityType
    from .workflow_execution import WorkflowExecution


class ActivityTask(BaseModel):
    def __init__(
        self,
        activity_id: str,
        activity_type: "ActivityType",
        scheduled_event_id: int,
        workflow_execution: "WorkflowExecution",
        timeouts: Dict[str, Any],
        workflow_input: Any = None,
    ):
        self.activity_id = activity_id
        self.activity_type = activity_type
        self.details = None
        self.input = workflow_input
        self.last_heartbeat_timestamp = unix_time()
        self.scheduled_event_id = scheduled_event_id
        self.started_event_id: Optional[int] = None
        self.state = "SCHEDULED"
        self.task_token = str(mock_random.uuid4())
        self.timeouts = timeouts
        self.timeout_type: Optional[str] = None
        self.workflow_execution = workflow_execution
        # this is *not* necessarily coherent with workflow execution history,
        # but that shouldn't be a problem for tests
        self.scheduled_at = utcnow()

    def _check_workflow_execution_open(self) -> None:
        if not self.workflow_execution.open:
            raise SWFWorkflowExecutionClosedError()

    @property
    def open(self) -> bool:
        return self.state in ["SCHEDULED", "STARTED"]

    def to_full_dict(self) -> Dict[str, Any]:
        hsh: Dict[str, Any] = {
            "activityId": self.activity_id,
            "activityType": self.activity_type.to_short_dict(),
            "taskToken": self.task_token,
            "startedEventId": self.started_event_id,
            "workflowExecution": self.workflow_execution.to_short_dict(),
        }
        if self.input:
            hsh["input"] = self.input
        return hsh

    def start(self, started_event_id: int) -> None:
        self.state = "STARTED"
        self.started_event_id = started_event_id

    def complete(self) -> None:
        self._check_workflow_execution_open()
        self.state = "COMPLETED"

    def fail(self) -> None:
        self._check_workflow_execution_open()
        self.state = "FAILED"

    def reset_heartbeat_clock(self) -> None:
        self.last_heartbeat_timestamp = unix_time()

    def first_timeout(self) -> Optional[Timeout]:
        if not self.open or not self.workflow_execution.open:
            return None

        if self.timeouts["heartbeatTimeout"] == "NONE":
            return None

        heartbeat_timeout_at = self.last_heartbeat_timestamp + int(
            self.timeouts["heartbeatTimeout"]
        )
        _timeout = Timeout(self, heartbeat_timeout_at, "HEARTBEAT")
        if _timeout.reached:
            return _timeout
        return None

    def process_timeouts(self) -> None:
        _timeout = self.first_timeout()
        if _timeout:
            self.timeout(_timeout)

    def timeout(self, _timeout: Timeout) -> None:
        self._check_workflow_execution_open()
        self.state = "TIMED_OUT"
        self.timeout_type = _timeout.kind



import weakref
from collections import defaultdict
from typing import Any, Dict, Iterator, List, Optional

from moto.core.utils import iso_8601_datetime_with_milliseconds, utcnow
from moto.utilities.utils import filter_resources, merge_multiple_dicts

from ..exceptions import InvalidParameterValueErrorPeeringAttachment
from ..utils import describe_tag_filter, random_transit_gateway_attachment_id
from .core import TaggedEC2Resource
from .vpc_peering_connections import PeeringConnectionStatus


class TransitGatewayAttachment(TaggedEC2Resource):
    def __init__(
        self,
        backend: Any,
        resource_id: str,
        resource_type: str,
        transit_gateway_id: str,
        tags: Optional[Dict[str, str]] = None,
    ):
        self.ec2_backend = backend
        self.association: Dict[str, str] = {}
        self.propagation: Dict[str, str] = {}
        self.resource_id = resource_id
        self.resource_type = resource_type

        self.id = random_transit_gateway_attachment_id()
        self.transit_gateway_id = transit_gateway_id

        self.state = "available"
        self.add_tags(tags or {})

        self._created_at = utcnow()
        self.resource_owner_id = backend.account_id
        self.transit_gateway_owner_id = backend.account_id
        self.owner_id = backend.account_id

    @property
    def create_time(self) -> str:
        return iso_8601_datetime_with_milliseconds(self._created_at)


class TransitGatewayVpcAttachment(TransitGatewayAttachment):
    DEFAULT_OPTIONS = {
        "ApplianceModeSupport": "disable",
        "DnsSupport": "enable",
        "Ipv6Support": "disable",
    }

    def __init__(
        self,
        backend: Any,
        transit_gateway_id: str,
        vpc_id: str,
        subnet_ids: List[str],
        tags: Optional[Dict[str, str]] = None,
        options: Optional[Dict[str, str]] = None,
    ):
        super().__init__(
            backend=backend,
            transit_gateway_id=transit_gateway_id,
            resource_id=vpc_id,
            resource_type="vpc",
            tags=tags,
        )

        self.vpc_id = vpc_id
        self.subnet_ids = subnet_ids
        self.options = merge_multiple_dicts(self.DEFAULT_OPTIONS, options or {})


class TransitGatewayPeeringAttachment(TransitGatewayAttachment):
    def __init__(
        self,
        backend: Any,
        transit_gateway_id: str,
        peer_transit_gateway_id: str,
        peer_region: str,
        peer_account_id: str,
        tags: Dict[str, str],
        region_name: str,
    ):
        super().__init__(
            backend=backend,
            transit_gateway_id=transit_gateway_id,
            resource_id=peer_transit_gateway_id,
            resource_type="peering",
            tags=tags,
        )

        self.accepter_tgw_info = {
            "ownerId": peer_account_id,
            "region": peer_region,
            "transitGatewayId": peer_transit_gateway_id,
        }
        self.requester_tgw_info = {
            "ownerId": self.owner_id,
            "region": region_name,
            "transitGatewayId": transit_gateway_id,
        }
        self.status = PeeringConnectionStatus(accepter_id=peer_account_id)


class TransitGatewayAttachmentBackend:
    backend_refs = defaultdict(set)  # type: ignore

    def __init__(self) -> None:
        self.transit_gateway_attachments: Dict[str, TransitGatewayAttachment] = {}
        self.backend_refs[self.__class__].add(weakref.ref(self))

    @classmethod
    def _get_peering_attachment_backend_refs(
        cls,
    ) -> Iterator["TransitGatewayAttachmentBackend"]:
        for backend_ref in cls.backend_refs[cls]:
            backend = backend_ref()
            if backend is not None:
                yield backend

    def create_transit_gateway_vpn_attachment(
        self,
        vpn_id: str,
        transit_gateway_id: str,
        tags: Optional[Dict[str, str]] = None,
    ) -> TransitGatewayAttachment:
        transit_gateway_vpn_attachment = TransitGatewayAttachment(
            self,
            resource_id=vpn_id,
            resource_type="vpn",
            transit_gateway_id=transit_gateway_id,
            tags=tags,
        )
        self.transit_gateway_attachments[transit_gateway_vpn_attachment.id] = (
            transit_gateway_vpn_attachment
        )
        return transit_gateway_vpn_attachment

    def create_transit_gateway_vpc_attachment(
        self,
        transit_gateway_id: str,
        vpc_id: str,
        subnet_ids: List[str],
        tags: Optional[Dict[str, str]] = None,
        options: Optional[Dict[str, str]] = None,
    ) -> TransitGatewayVpcAttachment:
        transit_gateway_vpc_attachment = TransitGatewayVpcAttachment(
            self,
            transit_gateway_id=transit_gateway_id,
            tags=tags,
            vpc_id=vpc_id,
            subnet_ids=subnet_ids,
            options=options,
        )
        self.transit_gateway_attachments[transit_gateway_vpc_attachment.id] = (
            transit_gateway_vpc_attachment
        )
        return transit_gateway_vpc_attachment

    def describe_transit_gateway_attachments(
        self,
        transit_gateways_attachment_ids: Optional[List[str]] = None,
        filters: Any = None,
    ) -> List[TransitGatewayAttachment]:
        transit_gateway_attachments = list(self.transit_gateway_attachments.values())

        attr_pairs = (
            ("resource-id", "resource_id"),
            ("resource-type", "resource_type"),
            ("transit-gateway-id", "transit_gateway_id"),
        )

        if (
            not transit_gateways_attachment_ids == []
            and transit_gateways_attachment_ids is not None
        ):
            transit_gateway_attachments = [
                transit_gateways_attachment
                for transit_gateways_attachment in transit_gateway_attachments
                if transit_gateways_attachment.id in transit_gateways_attachment_ids
            ]

        result = transit_gateway_attachments
        if filters:
            result = filter_resources(transit_gateway_attachments, filters, attr_pairs)
        return result

    def describe_transit_gateway_vpc_attachments(
        self,
        transit_gateways_attachment_ids: Optional[List[str]] = None,
        filters: Any = None,
    ) -> List[TransitGatewayAttachment]:
        transit_gateway_attachments = list(self.transit_gateway_attachments.values())

        attr_pairs = (
            ("state", "state"),
            ("transit-gateway-attachment-id", "id"),
            ("transit-gateway-id", "transit_gateway_id"),
            ("vpc-id", "resource_id"),
        )

        if (
            not transit_gateways_attachment_ids == []
            and transit_gateways_attachment_ids is not None
        ):
            transit_gateway_attachments = [
                transit_gateways_attachment
                for transit_gateways_attachment in transit_gateway_attachments
                if transit_gateways_attachment.id in transit_gateways_attachment_ids
            ]

        result = transit_gateway_attachments
        if filters:
            result = filter_resources(transit_gateway_attachments, filters, attr_pairs)
        return result

    def delete_transit_gateway_vpc_attachment(
        self, transit_gateway_attachment_id: str
    ) -> TransitGatewayAttachment:
        transit_gateway_attachment = self.transit_gateway_attachments.pop(
            transit_gateway_attachment_id
        )
        transit_gateway_attachment.state = "deleted"
        return transit_gateway_attachment

    def modify_transit_gateway_vpc_attachment(
        self,
        transit_gateway_attachment_id: str,
        add_subnet_ids: Optional[List[str]] = None,
        options: Optional[Dict[str, str]] = None,
        remove_subnet_ids: Optional[List[str]] = None,
    ) -> TransitGatewayAttachment:
        tgw_attachment = self.transit_gateway_attachments[transit_gateway_attachment_id]
        if remove_subnet_ids:
            tgw_attachment.subnet_ids = [  # type: ignore[attr-defined]
                id
                for id in tgw_attachment.subnet_ids  # type: ignore[attr-defined]
                if id not in remove_subnet_ids
            ]

        if options:
            tgw_attachment.options.update(options)  # type: ignore[attr-defined]

        if add_subnet_ids:
            for subnet_id in add_subnet_ids:
                tgw_attachment.subnet_ids.append(subnet_id)  # type: ignore[attr-defined]

        return tgw_attachment

    def set_attachment_association(
        self, transit_gateway_attachment_id: str, transit_gateway_route_table_id: str
    ) -> None:
        self.transit_gateway_attachments[transit_gateway_attachment_id].association = {
            "state": "associated",
            "transitGatewayRouteTableId": transit_gateway_route_table_id,
        }

    def unset_attachment_association(self, tgw_attach_id: str) -> None:
        self.transit_gateway_attachments[tgw_attach_id].association = {}

    def set_attachment_propagation(
        self, transit_gateway_attachment_id: str, transit_gateway_route_table_id: str
    ) -> None:
        self.transit_gateway_attachments[transit_gateway_attachment_id].propagation = {
            "state": "enabled",
            "transitGatewayRouteTableId": transit_gateway_route_table_id,
        }

    def unset_attachment_propagation(self, tgw_attach_id: str) -> None:
        self.transit_gateway_attachments[tgw_attach_id].propagation = {}

    def disable_attachment_propagation(
        self, transit_gateway_attachment_id: str
    ) -> None:
        self.transit_gateway_attachments[transit_gateway_attachment_id].propagation[
            "state"
        ] = "disabled"

    def create_transit_gateway_peering_attachment(
        self,
        transit_gateway_id: str,
        peer_transit_gateway_id: str,
        peer_region: str,
        peer_account_id: str,
        tags: Dict[str, str],
    ) -> TransitGatewayPeeringAttachment:
        transit_gateway_peering_attachment = TransitGatewayPeeringAttachment(
            self,
            transit_gateway_id=transit_gateway_id,
            peer_transit_gateway_id=peer_transit_gateway_id,
            peer_region=peer_region,
            peer_account_id=peer_account_id,
            tags=tags,
            region_name=self.region_name,  # type: ignore[attr-defined]
        )

        self.transit_gateway_attachments[transit_gateway_peering_attachment.id] = (
            transit_gateway_peering_attachment
        )

        # If the peer is not same as the current account or region, create attachment in peer backend
        if self.account_id != peer_account_id or self.region_name != peer_region:  # type: ignore[attr-defined]
            for backend in self._get_peering_attachment_backend_refs():
                if (
                    backend.account_id == peer_account_id  # type: ignore[attr-defined]
                    and backend.region_name == peer_region  # type: ignore[attr-defined]
                ):
                    backend.transit_gateway_attachments[
                        transit_gateway_peering_attachment.id
                    ] = transit_gateway_peering_attachment

        transit_gateway_peering_attachment.status.pending()
        transit_gateway_peering_attachment.state = "pendingAcceptance"
        return transit_gateway_peering_attachment

    def describe_transit_gateway_peering_attachments(
        self,
        transit_gateways_attachment_ids: Optional[List[str]] = None,
        filters: Any = None,
    ) -> List[TransitGatewayAttachment]:
        transit_gateway_attachments = list(self.transit_gateway_attachments.values())

        attr_pairs = (
            ("state", "state"),
            ("transit-gateway-attachment-id", "id"),
            ("local-owner-id", "requester_tgw_info", "ownerId"),
            ("remote-owner-id", "accepter_tgw_info", "ownerId"),
        )

        if transit_gateways_attachment_ids:
            transit_gateway_attachments = [
                transit_gateways_attachment
                for transit_gateways_attachment in transit_gateway_attachments
                if transit_gateways_attachment.id in transit_gateways_attachment_ids
            ]

        if filters:
            transit_gateway_attachments = filter_resources(
                transit_gateway_attachments, filters, attr_pairs
            )
            transit_gateway_attachments = describe_tag_filter(
                filters, transit_gateway_attachments
            )
        return transit_gateway_attachments

    def accept_transit_gateway_peering_attachment(
        self, transit_gateway_attachment_id: str
    ) -> TransitGatewayAttachment:
        transit_gateway_attachment = self.transit_gateway_attachments[
            transit_gateway_attachment_id
        ]

        requester_account_id = transit_gateway_attachment.requester_tgw_info["ownerId"]  # type: ignore[attr-defined]
        requester_region_name = transit_gateway_attachment.requester_tgw_info["region"]  # type: ignore[attr-defined]
        accepter_account_id = transit_gateway_attachment.accepter_tgw_info["ownerId"]  # type: ignore[attr-defined]
        accepter_region_name = transit_gateway_attachment.accepter_tgw_info["region"]  # type: ignore[attr-defined]

        # For cross-account peering, must be accepted by the accepter
        if (
            requester_account_id != accepter_account_id
            and self.account_id != accepter_account_id  # type: ignore[attr-defined]
        ):
            raise InvalidParameterValueErrorPeeringAttachment(
                "accept", transit_gateway_attachment_id
            )

        if (
            requester_region_name != accepter_region_name
            and self.region_name != accepter_region_name  # type: ignore[attr-defined]
        ):
            raise InvalidParameterValueErrorPeeringAttachment(
                "accept", transit_gateway_attachment_id
            )

        transit_gateway_attachment.state = "available"
        # Bit dodgy - we just assume that we act on a TransitGatewayPeeringAttachment
        # We could just as easily have another sub-class of TransitGatewayAttachment on our hands, which does not have a status-attribute
        transit_gateway_attachment.status.accept()  # type: ignore[attr-defined]
        return transit_gateway_attachment

    def reject_transit_gateway_peering_attachment(
        self, transit_gateway_attachment_id: str
    ) -> TransitGatewayAttachment:
        transit_gateway_attachment = self.transit_gateway_attachments[
            transit_gateway_attachment_id
        ]

        requester_account_id = transit_gateway_attachment.requester_tgw_info["ownerId"]  # type: ignore[attr-defined]
        requester_region_name = transit_gateway_attachment.requester_tgw_info["region"]  # type: ignore[attr-defined]
        accepter_account_id = transit_gateway_attachment.accepter_tgw_info["ownerId"]  # type: ignore[attr-defined]
        accepter_region_name = transit_gateway_attachment.requester_tgw_info["region"]  # type: ignore[attr-defined]

        if (
            requester_account_id != accepter_account_id
            and self.account_id != accepter_account_id  # type: ignore[attr-defined]
        ):
            raise InvalidParameterValueErrorPeeringAttachment(
                "reject", transit_gateway_attachment_id
            )

        if (
            requester_region_name != accepter_region_name
            and self.region_name != accepter_region_name  # type: ignore[attr-defined]
        ):
            raise InvalidParameterValueErrorPeeringAttachment(
                "reject", transit_gateway_attachment_id
            )

        transit_gateway_attachment.state = "rejected"
        transit_gateway_attachment.status.reject()  # type: ignore[attr-defined]
        return transit_gateway_attachment

    def delete_transit_gateway_peering_attachment(
        self, transit_gateway_attachment_id: str
    ) -> TransitGatewayAttachment:
        transit_gateway_attachment = self.transit_gateway_attachments[
            transit_gateway_attachment_id
        ]
        transit_gateway_attachment.state = "deleted"
        transit_gateway_attachment.status.deleted(deleter_id=self.account_id)  # type: ignore[attr-defined]
        return transit_gateway_attachment



import json
from typing import Any, Dict, List, Optional, Tuple

from moto.core.common_models import ConfigQueryModel
from moto.core.exceptions import InvalidNextTokenException
from moto.s3 import s3_backends
from moto.s3.models import S3Backend


class S3ConfigQuery(ConfigQueryModel[S3Backend]):
    def list_config_service_resources(
        self,
        account_id: str,
        resource_ids: Optional[List[str]],
        resource_name: Optional[str],
        limit: int,
        next_token: Optional[str],
        backend_region: Optional[str] = None,
        resource_region: Optional[str] = None,
        aggregator: Optional[Dict[str, Any]] = None,
    ) -> Tuple[List[Dict[str, Any]], Optional[str]]:
        # The resource_region only matters for aggregated queries as you can filter on bucket regions for them.
        # For other resource types, you would need to iterate appropriately for the backend_region.

        # Resource IDs are the same as S3 bucket names
        # For aggregation -- did we get both a resource ID and a resource name?
        if resource_ids and resource_name:
            # If the values are different, then return an empty list:
            if resource_name not in resource_ids:
                return [], None

        # If no filter was passed in for resource names/ids then return them all:
        if not resource_ids and not resource_name:
            bucket_list = list(self.backends[account_id]["global"].buckets.keys())

        else:
            # Match the resource name / ID:
            bucket_list = []
            filter_buckets = [resource_name] if resource_name else resource_ids

            for bucket in self.backends[account_id]["global"].buckets.keys():
                if bucket in filter_buckets:  # type: ignore
                    bucket_list.append(bucket)

        # Filter on the proper region if supplied:
        region_filter = backend_region or resource_region
        if region_filter:
            region_buckets = []

            for bucket in bucket_list:
                if (
                    self.backends[account_id]["global"].buckets[bucket].region_name
                    == region_filter
                ):
                    region_buckets.append(bucket)

            bucket_list = region_buckets

        if not bucket_list:
            return [], None

        # Pagination logic:
        sorted_buckets = sorted(bucket_list)
        new_token = None

        # Get the start:
        if not next_token:
            start = 0
        else:
            # Tokens for this moto feature is just the bucket name:
            # For OTHER non-global resource types, it's the region concatenated with the resource ID.
            if next_token not in sorted_buckets:
                raise InvalidNextTokenException()

            start = sorted_buckets.index(next_token)

        # Get the list of items to collect:
        bucket_list = sorted_buckets[start : (start + limit)]

        if len(sorted_buckets) > (start + limit):
            new_token = sorted_buckets[start + limit]

        return (
            [
                {
                    "type": "AWS::S3::Bucket",
                    "id": bucket,
                    "name": bucket,
                    "region": self.backends[account_id]["global"]
                    .buckets[bucket]
                    .region_name,
                }
                for bucket in bucket_list
            ],
            new_token,
        )

    def get_config_resource(
        self,
        account_id: str,
        resource_id: str,
        resource_name: Optional[str] = None,
        backend_region: Optional[str] = None,
        resource_region: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        # Get the bucket:
        bucket = self.backends[account_id]["global"].buckets.get(resource_id)

        if not bucket:
            return None

        # Are we filtering based on region?
        region_filter = backend_region or resource_region
        if region_filter and bucket.region_name != region_filter:
            return None

        # Are we also filtering on bucket name?
        if resource_name and bucket.name != resource_name:
            return None

        # Format the bucket to the AWS Config format:
        config_data = bucket.to_config_dict()

        # The 'configuration' field is also a JSON string:
        config_data["configuration"] = json.dumps(config_data["configuration"])

        # Supplementary config need all values converted to JSON strings if they are not strings already:
        for field, value in config_data["supplementaryConfiguration"].items():
            if not isinstance(value, str):
                config_data["supplementaryConfiguration"][field] = json.dumps(value)

        return config_data


s3_config_query = S3ConfigQuery(s3_backends)



from .responses import GreengrassResponse

url_bases = [
    r"https?://greengrass\.(.+)\.amazonaws.com",
]

url_paths = {
    "{0}/greengrass/definition/cores$": GreengrassResponse.method_dispatch(
        GreengrassResponse.core_definitions
    ),
    "{0}/greengrass/definition/cores/(?P<definition_id>[^/]+)/?$": GreengrassResponse.method_dispatch(
        GreengrassResponse.core_definition
    ),
    "{0}/greengrass/definition/cores/(?P<definition_id>[^/]+)/versions$": GreengrassResponse.method_dispatch(
        GreengrassResponse.core_definition_versions
    ),
    "{0}/greengrass/definition/cores/(?P<definition_id>[^/]+)/versions/(?P<definition_version_id>[^/]+)/?$": GreengrassResponse.method_dispatch(
        GreengrassResponse.core_definition_version
    ),
    "{0}/greengrass/definition/devices$": GreengrassResponse.method_dispatch(
        GreengrassResponse.device_definitions
    ),
    "{0}/greengrass/definition/devices/(?P<definition_id>[^/]+)/?$": GreengrassResponse.method_dispatch(
        GreengrassResponse.device_definition
    ),
    "{0}/greengrass/definition/devices/(?P<definition_id>[^/]+)/versions$": GreengrassResponse.method_dispatch(
        GreengrassResponse.device_definition_versions
    ),
    "{0}/greengrass/definition/devices/(?P<definition_id>[^/]+)/versions/(?P<definition_version_id>[^/]+)/?$": GreengrassResponse.method_dispatch(
        GreengrassResponse.device_definition_version
    ),
    "{0}/greengrass/definition/functions$": GreengrassResponse.method_dispatch(
        GreengrassResponse.function_definitions
    ),
    "{0}/greengrass/definition/functions/(?P<definition_id>[^/]+)/?$": GreengrassResponse.method_dispatch(
        GreengrassResponse.function_definition
    ),
    "{0}/greengrass/definition/functions/(?P<definition_id>[^/]+)/versions$": GreengrassResponse.method_dispatch(
        GreengrassResponse.function_definition_versions
    ),
    "{0}/greengrass/definition/functions/(?P<definition_id>[^/]+)/versions/(?P<definition_version_id>[^/]+)/?$": GreengrassResponse.method_dispatch(
        GreengrassResponse.function_definition_version
    ),
    "{0}/greengrass/definition/resources$": GreengrassResponse.method_dispatch(
        GreengrassResponse.resource_definitions
    ),
    "{0}/greengrass/definition/resources/(?P<definition_id>[^/]+)/?$": GreengrassResponse.method_dispatch(
        GreengrassResponse.resource_definition
    ),
    "{0}/greengrass/definition/resources/(?P<definition_id>[^/]+)/versions$": GreengrassResponse.method_dispatch(
        GreengrassResponse.resource_definition_versions
    ),
    "{0}/greengrass/definition/subscriptions$": GreengrassResponse.method_dispatch(
        GreengrassResponse.subscription_definitions
    ),
    "{0}/greengrass/definition/subscriptions/(?P<definition_id>[^/]+)/?$": GreengrassResponse.method_dispatch(
        GreengrassResponse.subscription_definition
    ),
    "{0}/greengrass/definition/subscriptions/(?P<definition_id>[^/]+)/versions$": GreengrassResponse.method_dispatch(
        GreengrassResponse.subscription_definition_versions
    ),
    "{0}/greengrass/definition/subscriptions/(?P<definition_id>[^/]+)/versions/(?P<definition_version_id>[^/]+)/?$": GreengrassResponse.method_dispatch(
        GreengrassResponse.subscription_definition_version
    ),
    "{0}/greengrass/definition/resources/(?P<definition_id>[^/]+)/versions/(?P<definition_version_id>[^/]+)/?$": GreengrassResponse.method_dispatch(
        GreengrassResponse.resource_definition_version
    ),
    "{0}/greengrass/groups$": GreengrassResponse.method_dispatch(
        GreengrassResponse.groups
    ),
    "{0}/greengrass/groups/(?P<group_id>[^/]+)/?$": GreengrassResponse.method_dispatch(
        GreengrassResponse.group
    ),
    "{0}/greengrass/groups/(?P<group_id>[^/]+)/role$": GreengrassResponse.method_dispatch(
        GreengrassResponse.role
    ),
    "{0}/greengrass/groups/(?P<group_id>[^/]+)/versions$": GreengrassResponse.method_dispatch(
        GreengrassResponse.group_versions
    ),
    "{0}/greengrass/groups/(?P<group_id>[^/]+)/deployments$": GreengrassResponse.method_dispatch(
        GreengrassResponse.deployments
    ),
    "{0}/greengrass/groups/(?P<group_id>[^/]+)/deployments/\\$reset$": GreengrassResponse.method_dispatch(
        GreengrassResponse.deployments_reset
    ),
    "{0}/greengrass/groups/(?P<group_id>[^/]+)/deployments/(?P<group_version_id>[^/]+)/status$": GreengrassResponse.method_dispatch(
        GreengrassResponse.deployment_satus
    ),
    "{0}/greengrass/groups/(?P<group_id>[^/]+)/versions/(?P<group_version_id>[^/]+)/?$": GreengrassResponse.method_dispatch(
        GreengrassResponse.group_version
    ),
}



"""
Neptune is a bit of an odd duck.
It shares almost everything with RDS: the endpoint URL, and the features. Only the parameters to these features can be different.

Because the endpoint URL is the same (rds.amazonaws.com), every request is intercepted by the RDS service.
RDS then has to determine whether any incoming call was meant for RDS, or for neptune.
"""

from .models import neptune_backends  # noqa: F401



from typing import Dict, Final, Optional, Set

from botocore.exceptions import ClientError

from moto.stepfunctions.parser.api import HistoryEventType, TaskFailedEventDetails
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ResourceRuntimePart,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_callback import (
    StateTaskServiceCallback,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.boto_client import boto_client_for
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str


class StateTaskServiceSns(StateTaskServiceCallback):
    _SUPPORTED_API_PARAM_BINDINGS: Final[Dict[str, Set[str]]] = {
        "publish": {
            "Message",
            "MessageAttributes",
            "MessageStructure",
            "PhoneNumber",
            "Subject",
            "TargetArn",
            "TopicArn",
        }
    }

    def _get_supported_parameters(self) -> Optional[Set[str]]:
        return self._SUPPORTED_API_PARAM_BINDINGS.get(self.resource.api_action.lower())

    def _from_error(self, env: Environment, ex: Exception) -> FailureEvent:
        if isinstance(ex, ClientError):
            error_code = ex.response["Error"]["Code"]

            exception_name = error_code
            if not exception_name.endswith("Exception"):
                exception_name += "Exception"
            error_name = f"SNS.{exception_name}"

            error_message = ex.response["Error"]["Message"]
            status_code = ex.response["ResponseMetadata"]["HTTPStatusCode"]
            request_id = ex.response["ResponseMetadata"]["RequestId"]
            error_cause = (
                f"{error_message} "
                f"(Service: AmazonSNS; "
                f"Status Code: {status_code}; "
                f"Error Code: {error_code}; "
                f"Request ID: {request_id}; "
                f"Proxy: null)"
            )

            return FailureEvent(
                error_name=CustomErrorName(error_name=error_name),
                event_type=HistoryEventType.TaskFailed,
                event_details=EventDetails(
                    taskFailedEventDetails=TaskFailedEventDetails(
                        error=error_name,
                        cause=error_cause,
                        resource=self._get_sfn_resource(),
                        resourceType=self._get_sfn_resource_type(),
                    )
                ),
            )
        return super()._from_error(env=env, ex=ex)

    def _eval_service_task(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
    ):
        service_name = self._get_boto_service_name()
        api_action = self._get_boto_service_action()
        sns_client = boto_client_for(
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            service=service_name,
        )

        # Optimised integration automatically stringifies
        if "Message" in normalised_parameters and not isinstance(
            message := normalised_parameters["Message"], str
        ):
            normalised_parameters["Message"] = to_json_str(message)

        response = getattr(sns_client, api_action)(**normalised_parameters)
        response.pop("ResponseMetadata", None)
        env.stack.append(response)



PAGINATION_MODEL = {
    "list_permission_sets": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "result_key": "PermissionSets",
        "unique_attribute": "permission_set_arn",
    },
    "list_account_assignments_for_principal": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "result_key": "AccountAssignments",
        "unique_attribute": [
            "AccountId",
            "PermissionSetArn",
            "PrincipalId",
            "PrincipalType",
        ],
    },
    "list_account_assignments": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "result_key": "AccountAssignments",
        "unique_attribute": [
            "AccountId",
            "PermissionSetArn",
            "PrincipalId",
            "PrincipalType",
        ],
    },
    "list_managed_policies_in_permission_set": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "result_key": "AttachedManagedPolicies",
        "unique_attribute": ["arn"],
    },
    "list_customer_managed_policy_references_in_permission_set": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "result_key": "CustomerManagedPolicyReferences",
        "unique_attribute": ["name", "path"],
    },
}



import json

from jsonpath_ng import parse

from moto.stepfunctions.parser.asl.utils.encoding import to_json_str


class JSONPathUtils:
    @staticmethod
    def extract_json(path: str, data: json) -> json:
        input_expr = parse(path)
        find_res = [match.value for match in input_expr.find(data)]
        if find_res == list():
            raise RuntimeError(
                f"The JSONPath {path} could not be found in the input {to_json_str(data)}"
            )
        if len(find_res) == 1:
            value = find_res[0]
        else:
            value = find_res
        return value



from typing import Any, Dict, List, Optional


class Policy(object):
    def __init__(self, policy_name: str, policy_type_name: str):
        self.policy_name = policy_name
        self.policy_type_name = policy_type_name


class AppCookieStickinessPolicy(Policy):
    def __init__(self, policy_name: str, cookie_name: str):
        super().__init__(policy_name, policy_type_name="AppCookieStickinessPolicy")
        self.cookie_name = cookie_name


class LbCookieStickinessPolicy(Policy):
    def __init__(self, policy_name: str, cookie_expiration_period: Optional[int]):
        super().__init__(policy_name, policy_type_name="LbCookieStickinessPolicy")
        self.cookie_expiration_period = cookie_expiration_period


class OtherPolicy(Policy):
    def __init__(
        self,
        policy_name: str,
        policy_type_name: str,
        policy_attrs: List[Dict[str, Any]],
    ):
        super().__init__(policy_name, policy_type_name=policy_type_name)
        self.attributes = policy_attrs or []



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

if TYPE_CHECKING:
    from py_partiql_parser import QueryMetadata


def query(
    statement: str, source_data: Dict[str, str], parameters: List[Dict[str, Any]]
) -> Tuple[
    List[Dict[str, Any]],
    Dict[str, List[Tuple[Optional[Dict[str, Any]], Optional[Dict[str, Any]]]]],
]:
    from py_partiql_parser import DynamoDBStatementParser

    return DynamoDBStatementParser(source_data).parse(statement, parameters)


def get_query_metadata(statement: str) -> "QueryMetadata":
    from py_partiql_parser import DynamoDBStatementParser

    return DynamoDBStatementParser.get_query_metadata(query=statement)



from .responses import ElasticMapReduceResponse

url_bases = [
    r"https?://(.+)\.elasticmapreduce\.amazonaws.com",
    r"https?://elasticmapreduce\.(.+)\.amazonaws.com",
]

url_paths = {"{0}/$": ElasticMapReduceResponse.dispatch}



"""emrserverless base URL and path."""

from .responses import EMRServerlessResponse

url_bases = [
    r"https?://emr-serverless\.(.+)\.amazonaws\.com",
]


url_paths = {
    "{0}/applications$": EMRServerlessResponse.dispatch,
    "{0}/applications/(?P<applicationId>[^/]+)$": EMRServerlessResponse.dispatch,
    "{0}/applications/(?P<applicationId>[^/]+)/start$": EMRServerlessResponse.dispatch,
    "{0}/applications/(?P<applicationId>[^/]+)/stop$": EMRServerlessResponse.dispatch,
    "{0}/applications/(?P<applicationId>[^/]+)/jobruns$": EMRServerlessResponse.dispatch,
    "{0}/applications/(?P<applicationId>[^/]+)/jobruns/(?P<jobRunId>[^/]+)$": EMRServerlessResponse.dispatch,
}



from .responses import ElasticTranscoderResponse

url_bases = [
    r"https?://elastictranscoder\.(.+)\.amazonaws.com",
]


url_paths = {
    r"{0}/(?P<api_version>[^/]+)/pipelines/?$": ElasticTranscoderResponse.method_dispatch(
        ElasticTranscoderResponse.pipelines
    ),
    r"{0}/(?P<api_version>[^/]+)/pipelines/(?P<pipeline_id>[^/]+)/?$": ElasticTranscoderResponse.method_dispatch(
        ElasticTranscoderResponse.individual_pipeline
    ),
}



from typing import Final, Optional

from moto.stepfunctions.parser.asl.component.common.comment import Comment
from moto.stepfunctions.parser.asl.component.common.flow.start_at import StartAt
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.processor_config import (
    ProcessorConfig,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iteration_declaration import (
    IterationDecl,
)
from moto.stepfunctions.parser.asl.component.states import States


class ItemProcessorDecl(IterationDecl):
    processor_config: Final[ProcessorConfig]

    def __init__(
        self,
        comment: Optional[Comment],
        start_at: StartAt,
        states: States,
        processor_config: Optional[ProcessorConfig],
    ):
        super().__init__(start_at=start_at, comment=comment, states=states)
        self.processor_config = processor_config or ProcessorConfig()



import base64
import hashlib
from datetime import datetime
from typing import Any


def deep_convert_datetime_to_isoformat(obj: Any) -> Any:
    if isinstance(obj, datetime):
        return obj.isoformat()
    elif isinstance(obj, list):
        return [deep_convert_datetime_to_isoformat(x) for x in obj]
    elif isinstance(obj, dict):
        return {k: deep_convert_datetime_to_isoformat(v) for k, v in obj.items()}
    else:
        return obj


def hash_device_name(name: str) -> str:
    digest = hashlib.md5(name.encode("utf-8")).digest()
    token = base64.b64encode(digest)
    return token.decode("utf-8")



from typing import Final, List

from moto.stepfunctions.parser.asl.component.common.catch.catch_outcome import (
    CatchOutcome,
)
from moto.stepfunctions.parser.asl.component.common.catch.catcher_decl import (
    CatcherDecl,
)
from moto.stepfunctions.parser.asl.component.common.catch.catcher_outcome import (
    CatcherOutcome,
    CatcherOutcomeCaught,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class CatchDecl(EvalComponent):
    def __init__(self, catchers: List[CatcherDecl]):
        self.catchers: Final[List[CatcherDecl]] = catchers

    def _eval_body(self, env: Environment) -> None:
        for catcher in self.catchers:
            catcher.eval(env)
            catcher_outcome: CatcherOutcome = env.stack.pop()

            if isinstance(catcher_outcome, CatcherOutcomeCaught):
                env.stack.append(CatchOutcome.Caught)
                return

        env.stack.append(CatchOutcome.NotCaught)



from typing import Final


class AWSExecutionDetails:
    account: Final[str]
    region: Final[str]
    role_arn: Final[str]

    def __init__(self, account: str, region: str, role_arn: str):
        self.account = account
        self.region = region
        self.role_arn = role_arn



from typing import Final, Optional

from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.utils.json_path import JSONPathUtils


class OutputPath(EvalComponent):
    DEFAULT_PATH: Final[str] = "$"

    output_path: Final[Optional[str]]

    def __init__(self, output_path: Optional[str]):
        self.output_path = output_path

    def _eval_body(self, env: Environment) -> None:
        if self.output_path is None:
            env.inp = dict()
        else:
            current_output = env.stack.pop()
            state_output = JSONPathUtils.extract_json(self.output_path, current_output)
            env.inp = state_output
